#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>

/* .h4m (HVQM4 1.5) audio decoder 0.0 by hcs */

/* big endian */

static uint8_t get8(FILE *infile)
{
    uint8_t buf[1];
    if (1 != fread(buf, 1, 1, infile))
    {
        fprintf(stderr, "read error at 0x%lx\n", (unsigned long)ftell(infile));
        exit(EXIT_FAILURE);
    }

    return buf[0];
}

static uint16_t read16(uint8_t * buf)
{
    uint32_t v = 0;
    for (int i = 0; i < 2; i++)
    {
        v <<= 8;
        v |= buf[i];
    }
    return v;
}

static uint16_t get16(FILE * infile)
{
    uint8_t buf[2];
    if (2 != fread(buf, 1, 2, infile))
    {
        fprintf(stderr, "read error at 0x%lx\n", (unsigned long)ftell(infile));
        exit(EXIT_FAILURE);
    }
    return read16(buf);
}

static uint32_t read32(uint8_t * buf)
{
    uint32_t v = 0;
    for (int i = 0; i < 4; i++)
    {
        v <<= 8;
        v |= buf[i];
    }
    return v;
}

static uint32_t get32(FILE * infile)
{
    uint8_t buf[4];
    if (4 != fread(buf, 1, 4, infile))
    {
        fprintf(stderr, "read error at 0x%lx\n", (unsigned long)ftell(infile));
        exit(EXIT_FAILURE);
    }
    return read32(buf);
}

static void expect32(uint32_t expected, FILE * infile)
{
    uint32_t v = get32(infile);
    if (v != expected)
    {
        fprintf(stderr, "expected 0x%08"PRIx32" at 0x%lx, got 0x%08"PRIx32"\n",
            expected, ftell(infile)-4, v);
        exit(EXIT_FAILURE);
    }
}

static void expect32_imm(uint32_t expected, uint32_t actual, unsigned long offset)
{
    if (expected != actual)
    {
        fprintf(stderr, "expected 0x%08lx to be 0x%08"PRIx32", got 0x%08"PRIx32"\n",
            offset, expected, actual);
        exit(EXIT_FAILURE);
    }
}

static void expect32_text(uint32_t expected, uint32_t actual, char *name)
{
    if (expected != actual)
    {
        fprintf(stderr, "expected %s to be 0x%08"PRIx32", got 0x%08"PRIx32"\n",
            name, expected, actual);
        exit(EXIT_FAILURE);
    }
}

static void seek_past(uint32_t offset, FILE * infile)
{
    if (-1 == fseek(infile, offset, SEEK_CUR))
    {
        fprintf(stderr, "seek by 0x%x failed\n", offset);
        exit(EXIT_FAILURE);
    }
}

/* audio decode */

struct audio_state
{
    struct
    {
        int16_t hist;
        uint8_t idx;
    } ch[2];
};

uint32_t step_size[89][8]={
    {0x0,0x1,0x3,0x4,0x7,0x8,0xa,0xb}, /*0*/
    {0x1,0x3,0x5,0x7,0x9,0xb,0xd,0xf}, /*1*/
    {0x1,0x3,0x5,0x7,0xa,0xc,0xe,0x10}, /*2*/
    {0x1,0x3,0x6,0x8,0xb,0xd,0x10,0x12}, /*3*/
    {0x1,0x3,0x6,0x8,0xc,0xe,0x11,0x13}, /*4*/
    {0x1,0x4,0x7,0xa,0xd,0x10,0x13,0x16}, /*5*/
    {0x1,0x4,0x7,0xa,0xe,0x11,0x14,0x17}, /*6*/
    {0x1,0x4,0x8,0xb,0xf,0x12,0x16,0x19}, /*7*/
    {0x2,0x6,0xa,0xe,0x12,0x16,0x1a,0x1e}, /*8*/
    {0x2,0x6,0xa,0xe,0x13,0x17,0x1b,0x1f}, /*9*/
    {0x2,0x6,0xb,0xf,0x15,0x19,0x1e,0x22}, /*10*/
    {0x2,0x7,0xc,0x11,0x17,0x1c,0x21,0x26}, /*11*/
    {0x2,0x7,0xd,0x12,0x19,0x1e,0x24,0x29}, /*12*/
    {0x3,0x9,0xf,0x15,0x1c,0x22,0x28,0x2e}, /*13*/
    {0x3,0xa,0x11,0x18,0x1f,0x26,0x2d,0x34}, /*14*/
    {0x3,0xa,0x12,0x19,0x22,0x29,0x31,0x38}, /*15*/
    {0x4,0xc,0x15,0x1d,0x26,0x2e,0x37,0x3f}, /*16*/
    {0x4,0xd,0x16,0x1f,0x29,0x32,0x3b,0x44}, /*17*/
    {0x5,0xf,0x19,0x23,0x2e,0x38,0x42,0x4c}, /*18*/
    {0x5,0x10,0x1b,0x26,0x32,0x3d,0x48,0x53}, /*19*/
    {0x6,0x12,0x1f,0x2b,0x38,0x44,0x51,0x5d}, /*20*/
    {0x6,0x13,0x21,0x2e,0x3d,0x4a,0x58,0x65}, /*21*/
    {0x7,0x16,0x25,0x34,0x43,0x52,0x61,0x70}, /*22*/
    {0x8,0x18,0x29,0x39,0x4a,0x5a,0x6b,0x7b}, /*23*/
    {0x9,0x1b,0x2d,0x3f,0x52,0x64,0x76,0x88}, /*24*/
    {0xa,0x1e,0x32,0x46,0x5a,0x6e,0x82,0x96}, /*25*/
    {0xb,0x21,0x37,0x4d,0x63,0x79,0x8f,0xa5}, /*26*/
    {0xc,0x24,0x3c,0x54,0x6d,0x85,0x9d,0xb5}, /*27*/
    {0xd,0x27,0x42,0x5c,0x78,0x92,0xad,0xc7}, /*28*/
    {0xe,0x2b,0x49,0x66,0x84,0xa1,0xbf,0xdc}, /*29*/
    {0x10,0x30,0x51,0x71,0x92,0xb2,0xd3,0xf3}, /*30*/
    {0x11,0x34,0x58,0x7b,0xa0,0xc3,0xe7,0x10a}, /*31*/
    {0x13,0x3a,0x61,0x88,0xb0,0xd7,0xfe,0x125}, /*32*/
    {0x15,0x40,0x6b,0x96,0xc2,0xed,0x118,0x143}, /*33*/
    {0x17,0x46,0x76,0xa5,0xd5,0x104,0x134,0x163}, /*34*/
    {0x1a,0x4e,0x82,0xb6,0xeb,0x11f,0x153,0x187}, /*35*/
    {0x1c,0x55,0x8f,0xc8,0x102,0x13b,0x175,0x1ae}, /*36*/
    {0x1f,0x5e,0x9d,0xdc,0x11c,0x15b,0x19a,0x1d9}, /*37*/
    {0x22,0x67,0xad,0xf2,0x139,0x17e,0x1c4,0x209}, /*38*/
    {0x26,0x72,0xbf,0x10b,0x159,0x1a5,0x1f2,0x23e}, /*39*/
    {0x2a,0x7e,0xd2,0x126,0x17b,0x1cf,0x223,0x277}, /*40*/
    {0x2e,0x8a,0xe7,0x143,0x1a1,0x1fd,0x25a,0x2b6}, /*41*/
    {0x33,0x99,0xff,0x165,0x1cb,0x231,0x297,0x2fd}, /*42*/
    {0x38,0xa8,0x118,0x188,0x1f9,0x269,0x2d9,0x349}, /*43*/
    {0x3d,0xb8,0x134,0x1af,0x22b,0x2a6,0x322,0x39d}, /*44*/
    {0x44,0xcc,0x154,0x1dc,0x264,0x2ec,0x374,0x3fc}, /*45*/
    {0x4a,0xdf,0x175,0x20a,0x2a0,0x335,0x3cb,0x460}, /*46*/
    {0x52,0xf6,0x19b,0x23f,0x2e4,0x388,0x42d,0x4d1}, /*47*/
    {0x5a,0x10f,0x1c4,0x279,0x32e,0x3e3,0x498,0x54d}, /*48*/
    {0x63,0x12a,0x1f1,0x2b8,0x37f,0x446,0x50d,0x5d4}, /*49*/
    {0x6d,0x148,0x223,0x2fe,0x3d9,0x4b4,0x58f,0x66a}, /*50*/
    {0x78,0x168,0x259,0x349,0x43b,0x52b,0x61c,0x70c}, /*51*/
    {0x84,0x18d,0x296,0x39f,0x4a8,0x5b1,0x6ba,0x7c3}, /*52*/
    {0x91,0x1b4,0x2d8,0x3fb,0x51f,0x642,0x766,0x889}, /*53*/
    {0xa0,0x1e0,0x321,0x461,0x5a2,0x6e2,0x823,0x963}, /*54*/
    {0xb0,0x210,0x371,0x4d1,0x633,0x793,0x8f4,0xa54}, /*55*/
    {0xc2,0x246,0x3ca,0x54e,0x6d2,0x856,0x9da,0xb5e}, /*56*/
    {0xd5,0x27f,0x42a,0x5d4,0x780,0x92a,0xad5,0xc7f}, /*57*/
    {0xea,0x2bf,0x495,0x66a,0x840,0xa15,0xbeb,0xdc0}, /*58*/
    {0x102,0x306,0x50b,0x70f,0x914,0xb18,0xd1d,0xf21}, /*59*/
    {0x11c,0x354,0x58c,0x7c4,0x9fc,0xc34,0xe6c,0x10a4}, /*60*/
    {0x138,0x3a8,0x619,0x889,0xafb,0xd6b,0xfdc,0x124c}, /*61*/
    {0x157,0x406,0x6b5,0x964,0xc14,0xec3,0x1172,0x1421}, /*62*/
    {0x17a,0x46e,0x762,0xa56,0xd4a,0x103e,0x1332,0x1626}, /*63*/
    {0x19f,0x4de,0x81e,0xb5d,0xe9e,0x11dd,0x151d,0x185c}, /*64*/
    {0x1c9,0x55c,0x8ef,0xc82,0x1015,0x13a8,0x173b,0x1ace}, /*65*/
    {0x1f7,0x5e5,0x9d4,0xdc2,0x11b1,0x159f,0x198e,0x1d7c}, /*66*/
    {0x229,0x67c,0xacf,0xf22,0x1375,0x17c8,0x1c1b,0x206e}, /*67*/
    {0x260,0x721,0xbe3,0x10a4,0x1567,0x1a28,0x1eea,0x23ab}, /*68*/
    {0x29d,0x7d8,0xd14,0x124f,0x178b,0x1cc6,0x2202,0x273d}, /*69*/
    {0x2e0,0x8a1,0xe63,0x1424,0x19e6,0x1fa7,0x2569,0x2b2a}, /*70*/
    {0x32a,0x97f,0xfd4,0x1629,0x1c7e,0x22d3,0x2928,0x2f7d}, /*71*/
    {0x37b,0xa72,0x1169,0x1860,0x1f57,0x264e,0x2d45,0x343c}, /*72*/
    {0x3d4,0xb7d,0x1326,0x1acf,0x2279,0x2a22,0x31cb,0x3974}, /*73*/
    {0x436,0xca3,0x1511,0x1d7e,0x25ec,0x2e59,0x36c7,0x3f34}, /*74*/
    {0x4a2,0xde7,0x172c,0x2071,0x29b7,0x32fc,0x3c41,0x4586}, /*75*/
    {0x519,0xf4b,0x197e,0x23b0,0x2de3,0x3815,0x4248,0x4c7a}, /*76*/
    {0x59b,0x10d2,0x1c0a,0x2741,0x327a,0x3db1,0x48e9,0x5420}, /*77*/
    {0x62b,0x1281,0x1ed8,0x2b2e,0x3786,0x43dc,0x5033,0x5c89}, /*78*/
    {0x6c9,0x145b,0x21ee,0x2f80,0x3d14,0x4aa6,0x5839,0x65cb}, /*79*/
    {0x777,0x1665,0x2553,0x3441,0x4330,0x521e,0x610c,0x6ffa}, /*80*/
    {0x836,0x18a2,0x290f,0x397b,0x49e8,0x5a54,0x6ac1,0x7b2d}, /*81*/
    {0x908,0x1b19,0x2d2a,0x3f3b,0x514c,0x635d,0x756e,0x877f}, /*82*/
    {0x9ef,0x1dce,0x31ae,0x458d,0x596d,0x6d4c,0x812c,0x950b}, /*83*/
    {0xaee,0x20ca,0x36a6,0x4c82,0x625f,0x783b,0x8e17,0xa3f3}, /*84*/
    {0xc05,0x2410,0x3c1c,0x5427,0x6c34,0x843f,0x9c4b,0xb456}, /*85*/
    {0xd39,0x27ac,0x4220,0x5c93,0x7707,0x917a,0xabee,0xc661}, /*86*/
    {0xe8c,0x2ba4,0x48bd,0x65d5,0x82ee,0xa006,0xbd1f,0xda37}, /*87*/
    {0xfff,0x2ffe,0x4ffe,0x6ffd,0x8ffe,0xaffd,0xcffd,0xeffc}, /*88*/
};
uint8_t next_index[89][8]={
    {0,0,0,0,2,4,6,8}, /*0*/
    {0,0,0,0,3,5,7,9}, /*1*/
    {1,1,1,1,4,6,8,10}, /*2*/
    {2,2,2,2,5,7,9,11}, /*3*/
    {3,3,3,3,6,8,10,12}, /*4*/
    {4,4,4,4,7,9,11,13}, /*5*/
    {5,5,5,5,8,10,12,14}, /*6*/
    {6,6,6,6,9,11,13,15}, /*7*/
    {7,7,7,7,10,12,14,16}, /*8*/
    {8,8,8,8,11,13,15,17}, /*9*/
    {9,9,9,9,12,14,16,18}, /*10*/
    {10,10,10,10,13,15,17,19}, /*11*/
    {11,11,11,11,14,16,18,20}, /*12*/
    {12,12,12,12,15,17,19,21}, /*13*/
    {13,13,13,13,16,18,20,22}, /*14*/
    {14,14,14,14,17,19,21,23}, /*15*/
    {15,15,15,15,18,20,22,24}, /*16*/
    {16,16,16,16,19,21,23,25}, /*17*/
    {17,17,17,17,20,22,24,26}, /*18*/
    {18,18,18,18,21,23,25,27}, /*19*/
    {19,19,19,19,22,24,26,28}, /*20*/
    {20,20,20,20,23,25,27,29}, /*21*/
    {21,21,21,21,24,26,28,30}, /*22*/
    {22,22,22,22,25,27,29,31}, /*23*/
    {23,23,23,23,26,28,30,32}, /*24*/
    {24,24,24,24,27,29,31,33}, /*25*/
    {25,25,25,25,28,30,32,34}, /*26*/
    {26,26,26,26,29,31,33,35}, /*27*/
    {27,27,27,27,30,32,34,36}, /*28*/
    {28,28,28,28,31,33,35,37}, /*29*/
    {29,29,29,29,32,34,36,38}, /*30*/
    {30,30,30,30,33,35,37,39}, /*31*/
    {31,31,31,31,34,36,38,40}, /*32*/
    {32,32,32,32,35,37,39,41}, /*33*/
    {33,33,33,33,36,38,40,42}, /*34*/
    {34,34,34,34,37,39,41,43}, /*35*/
    {35,35,35,35,38,40,42,44}, /*36*/
    {36,36,36,36,39,41,43,45}, /*37*/
    {37,37,37,37,40,42,44,46}, /*38*/
    {38,38,38,38,41,43,45,47}, /*39*/
    {39,39,39,39,42,44,46,48}, /*40*/
    {40,40,40,40,43,45,47,49}, /*41*/
    {41,41,41,41,44,46,48,50}, /*42*/
    {42,42,42,42,45,47,49,51}, /*43*/
    {43,43,43,43,46,48,50,52}, /*44*/
    {44,44,44,44,47,49,51,53}, /*45*/
    {45,45,45,45,48,50,52,54}, /*46*/
    {46,46,46,46,49,51,53,55}, /*47*/
    {47,47,47,47,50,52,54,56}, /*48*/
    {48,48,48,48,51,53,55,57}, /*49*/
    {49,49,49,49,52,54,56,58}, /*50*/
    {50,50,50,50,53,55,57,59}, /*51*/
    {51,51,51,51,54,56,58,60}, /*52*/
    {52,52,52,52,55,57,59,61}, /*53*/
    {53,53,53,53,56,58,60,62}, /*54*/
    {54,54,54,54,57,59,61,63}, /*55*/
    {55,55,55,55,58,60,62,64}, /*56*/
    {56,56,56,56,59,61,63,65}, /*57*/
    {57,57,57,57,60,62,64,66}, /*58*/
    {58,58,58,58,61,63,65,67}, /*59*/
    {59,59,59,59,62,64,66,68}, /*60*/
    {60,60,60,60,63,65,67,69}, /*61*/
    {61,61,61,61,64,66,68,70}, /*62*/
    {62,62,62,62,65,67,69,71}, /*63*/
    {63,63,63,63,66,68,70,72}, /*64*/
    {64,64,64,64,67,69,71,73}, /*65*/
    {65,65,65,65,68,70,72,74}, /*66*/
    {66,66,66,66,69,71,73,75}, /*67*/
    {67,67,67,67,70,72,74,76}, /*68*/
    {68,68,68,68,71,73,75,77}, /*69*/
    {69,69,69,69,72,74,76,78}, /*70*/
    {70,70,70,70,73,75,77,79}, /*71*/
    {71,71,71,71,74,76,78,80}, /*72*/
    {72,72,72,72,75,77,79,81}, /*73*/
    {73,73,73,73,76,78,80,82}, /*74*/
    {74,74,74,74,77,79,81,83}, /*75*/
    {75,75,75,75,78,80,82,84}, /*76*/
    {76,76,76,76,79,81,83,85}, /*77*/
    {77,77,77,77,80,82,84,86}, /*78*/
    {78,78,78,78,81,83,85,87}, /*79*/
    {79,79,79,79,82,84,86,88}, /*80*/
    {80,80,80,80,83,85,87,88}, /*81*/
    {81,81,81,81,84,86,88,88}, /*82*/
    {82,82,82,82,85,87,88,88}, /*83*/
    {83,83,83,83,86,88,88,88}, /*84*/
    {84,84,84,84,87,88,88,88}, /*85*/
    {85,85,85,85,88,88,88,88}, /*86*/
    {86,86,86,86,88,88,88,88}, /*87*/
    {87,87,87,87,88,88,88,88}, /*88*/
};

static int16_t clamp16(int32_t v)
{
    if (v > INT16_MAX) return INT16_MAX;
    else if (v < INT16_MIN) return INT16_MIN;
    return v;
}

static void decode_audio(struct audio_state *state, int first_aud, uint32_t sample_count, FILE *infile, FILE *outfile)
{
    int16_t * samples = malloc(sample_count*sizeof(int16_t)*2);
    uint32_t t = 0, i = 0;
    if (first_aud)
    {
        state->ch[1].hist = get8(infile);
        state->ch[1].hist <<= 8;
        uint8_t b = get8(infile);
        state->ch[1].hist |= (b & 0x80);
        state->ch[1].idx = b & 0x7f;

        state->ch[0].hist = get8(infile);
        state->ch[0].hist <<= 8;
        b = get8(infile);
        state->ch[0].hist |= (b & 0x80);
        state->ch[0].idx = b & 0x7f;

        if (state->ch[0].idx > 88 || state->ch[1].idx > 88)
        {
            fprintf(stderr, "invalid step index at 0x%lx\n", ftell(infile));
            exit(EXIT_FAILURE);
        }

        samples[t++] = state->ch[0].hist;
        samples[t++] = state->ch[1].hist;
        i ++;
    }

    for (; i < sample_count; i++)
    {
        uint8_t b = get8(infile);
        uint8_t sample = b & 7;
        int32_t delta = step_size[state->ch[0].idx][sample];
        if (b & 8) state->ch[0].hist = clamp16(state->ch[0].hist - delta);
        else state->ch[0].hist = clamp16(state->ch[0].hist + delta);
        state->ch[0].idx = next_index[state->ch[0].idx][sample];
        samples[t++] = state->ch[0].hist;

        sample = (b & 0x70)>>4;
        delta = step_size[state->ch[1].idx][sample];
        if (b & 0x80) state->ch[1].hist = clamp16(state->ch[1].hist - delta);
        else state->ch[1].hist = clamp16(state->ch[1].hist + delta);
        state->ch[1].idx = next_index[state->ch[1].idx][sample];
        samples[t++] = state->ch[1].hist;
    }

    if (sample_count != fwrite(samples, sizeof(int16_t)*2, sample_count, outfile))
    {
        fprintf(stderr, "error writing output\n");
        exit(EXIT_FAILURE);
    }

    free(samples);
}

/* stream structure */

struct HVQM4_15_header
{
    uint32_t header_size;   /* 0x10-0x13 */
    uint32_t body_size;     /* 0x14-0x17 */
    uint32_t blocks;        /* 0x18-0x1B */
    uint32_t audio_frames;  /* 0x1C-0x1F */
    uint32_t video_frames;  /* 0x20-0x23 */
    uint32_t unk24;         /* 0x24-0x27 (0x8257) */
    uint32_t unk28;         /* 0x28-0x2B (0x24e48) */
    uint32_t unk2C;         /* 0x2C-0x2F (0) */
    uint32_t audio_frame_sz;/* 0x30-0x33 */
    uint16_t hres;          /* 0x34-0x35 */
    uint16_t vres;          /* 0x36-0x37 */
    uint32_t unk38;         /* 0x38-0x3B (0x02020000) */
    uint32_t unk3C;         /* 0x3C-0x3F (0x02100000) */
    uint32_t audio_srate;   /* 0x40-0x43 */
};

static void load_header(struct HVQM4_15_header *header, uint8_t *raw_header)
{
    header->header_size = read32(&raw_header[0x10]);
    header->body_size = read32(&raw_header[0x14]);
    header->blocks = read32(&raw_header[0x18]);
    header->audio_frames = read32(&raw_header[0x1C]);
    header->video_frames = read32(&raw_header[0x20]);
    header->unk24 = read32(&raw_header[0x24]);
    header->unk28 = read32(&raw_header[0x28]);
    header->unk2C = read32(&raw_header[0x2C]);
    header->audio_frame_sz = read32(&raw_header[0x30]);
    header->hres = read16(&raw_header[0x34]);
    header->vres = read16(&raw_header[0x36]);
    header->unk38 = read32(&raw_header[0x38]);
    header->unk3C = read32(&raw_header[0x3C]);
    header->audio_srate = read32(&raw_header[0x40]);

    expect32_text(0x44, header->header_size, "header size");
    /* no check for body size yet */
    /* no check for block, audio, video frame count */
    expect32_imm(0x8257, header->unk24, 0x24);
    expect32_imm(0x24e48, header->unk28, 0x28);
    expect32_imm(0, header->unk2C, 0x2C);
    expect32_text(0x650, header->audio_frame_sz, "audio frame size");
    /* no check for hres, vres */
    expect32_imm(0x02020000, header->unk38, 0x38);
    expect32_imm(0x02100000, header->unk3C, 0x3C);
    /* no check for srate, can be 0 */
}

void display_header(struct HVQM4_15_header *header)
{
    printf("Header size: 0x%"PRIx32"\n", header->header_size);
    printf("Body size:   0x%"PRIx32"\n", header->body_size);
    printf("Resolution:  %"PRIu32"x%"PRIu32"\n", header->hres, header->vres);
    printf("%d Blocks\n", header->blocks);
    printf("%d Video frames\n", header->video_frames);
    if (header->audio_frames)
    {
        printf("%d Audio frames\n", header->audio_frames);
        printf("Sample rate: %"PRIu32" Hz\n", header->audio_srate);
        //printf("Audio frame size: 0x%"PRIx32"\n", header->audio_frame_sz);
    } else {
        printf("No audio!\n");
    }
    printf("\n");
}

static void put_32bitLE(uint8_t * buf, uint32_t v)
{
    for (unsigned int i = 0; i < 4; i++)
    {
        buf[i] = v & 0xFF;
        v >>= 8;
    }
}

static void put_16bitLE(uint8_t * buf, uint16_t v)
{
    for (unsigned int i = 0; i < 2; i++)
    {
        buf[i] = v & 0xFF;
        v >>= 8;
    }
}

/* make a header for 16-bit PCM .wav */
/* buffer must be 0x2c bytes */
static void make_wav_header(uint8_t * buf, int32_t sample_count, int32_t sample_rate, int channels) {
    size_t bytecount;

    bytecount = sample_count*channels*2;

    /* RIFF header */
    memcpy(buf+0, "RIFF", 4);
    /* size of RIFF */
    put_32bitLE(buf+4, (int32_t)(bytecount+0x2c-8));

    /* WAVE header */
    memcpy(buf+8, "WAVE", 4);

    /* WAVE fmt chunk */
    memcpy(buf+0xc, "fmt ", 4);
    /* size of WAVE fmt chunk */
    put_32bitLE(buf+0x10, 0x10);

    /* compression code 1=PCM */
    put_16bitLE(buf+0x14, 1);

    /* channel count */
    put_16bitLE(buf+0x16, channels);

    /* sample rate */
    put_32bitLE(buf+0x18, sample_rate);

    /* bytes per second */
    put_32bitLE(buf+0x1c, sample_rate*channels*2);

    /* block align */
    put_16bitLE(buf+0x20, (int16_t)(channels*2));

    /* significant bits per sample */
    put_16bitLE(buf+0x22, 2*8);

    /* PCM has no extra format bytes, so we don't even need to specify a count */

    /* WAVE data chunk */
    memcpy(buf+0x24, "data", 4);
    /* size of WAVE data chunk */
    put_32bitLE(buf+0x28, (int32_t)bytecount);
}

int main(int argc, char **argv)
{
    printf("h4m 'HVQM4 1.5' audio decoder 0.0 by hcs\n\n");
    if (argc != 3)
    {
        fprintf(stderr, "usage: %s file.h4m output.wav\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    /* open input */
    FILE *infile = fopen(argv[1], "rb");
    if (!infile)
    {
        fprintf(stderr, "failed opening %s\n", argv[1]);
        exit(EXIT_FAILURE);
    }

    /* read in header */
    uint8_t raw_header[0x44];
    if (0x44 != fread(&raw_header, 1, 0x44, infile))
    {
        fprintf(stderr, "failed reading header");
        exit(EXIT_FAILURE);
    }

    /* check MAGIC */
    const char HVQM4_15_magic[16] = "HVQM4 1.5";
    if (memcmp(HVQM4_15_magic,&raw_header[0],16))
    {
        fprintf(stderr, "does not appear to be a HVQM4 1.5 file\n");
        exit(EXIT_FAILURE);
    }

    /* load up and check header */
    struct HVQM4_15_header header;
    load_header(&header, raw_header);
    display_header(&header);

    if (header.audio_frames == 0)
    {
        fprintf(stderr, "this video contains no audio!\n");
        exit(EXIT_FAILURE);
    }

    /* open output */
    FILE *outfile = fopen(argv[2], "wb");
    if (!outfile)
    {
        fprintf(stderr, "error opening %s\n", argv[2]);
        exit(EXIT_FAILURE);
    }

    /* fill in the space that we'll put the header in later */
    uint8_t riff_header[0x2c];
    if (0x2c != fwrite(riff_header, 1, 0x2c, outfile))
    {
        fprintf(stderr, "error writing riff header\n");
        exit(EXIT_FAILURE);
    }
    
    /* parse blocks */
    uint32_t block_count = 0;
    uint32_t total_aud_frames = 0;
    uint32_t total_vid_frames = 0;
    uint32_t last_block_start = ftell(infile);
    uint32_t total_sample_count = 0;
    while (block_count < header.blocks)
    {
        const long block_start = ftell(infile);
        expect32(ftell(infile) - last_block_start, infile);
        last_block_start = block_start;
        const uint32_t expected_block_size = get32(infile);
        const uint32_t expected_aud_frame_count = get32(infile);
        const uint32_t expected_vid_frame_count = get32(infile);
        expect32(0x01000000, infile);   /* EOS marker? */
        const long data_start = ftell(infile);

        block_count ++;
        //printf("block %d starts at 0x%lx, length 0x%"PRIx32"\n", (int)block_count, block_start, expected_block_size);

        /* parse frames */
        struct audio_state audio_state;
        int first_vid=1, first_aud=1;
        uint32_t vid_frame_count = 0, aud_frame_count = 0;
        int block_sample_count =0;
        while (aud_frame_count < expected_aud_frame_count ||
               vid_frame_count < expected_vid_frame_count)
        {
            const uint16_t frame_id1 = get16(infile);
            const uint16_t frame_id2 = get16(infile);
            const uint32_t frame_size = get32(infile);

            if (frame_id1 == 1 && (
                        (first_vid && frame_id2 == 0x10) ||
                        (!first_vid && frame_id2 == 0x20)))
            {
                /* video */
                first_vid = 0;
                vid_frame_count ++;
                total_vid_frames ++;
                //printf("video frame %d/%d (%d)\n", (int)vid_frame_count, (int)expected_vid_frame_count, (int)total_vid_frames);
                seek_past(frame_size, infile);
            }
            else if (frame_id1 == 0 && (
                        (first_aud && frame_id2 == 1) ||
                        (!first_aud && frame_id2 == 2)))
            {
                /* audio */
                const long audio_started = ftell(infile);
                const uint32_t samples = get32(infile);
                //decode_audio(&audio_state, first_aud, samples, infile, outfile);
                decode_audio(&audio_state, first_aud, samples, infile, outfile);
                block_sample_count += samples;
                aud_frame_count ++;
                total_aud_frames ++;
                //printf("0x%lx: audio frame %d/%d (%d) (%d samples)\n", (unsigned long)audio_started, (int)aud_frame_count, (int)expected_aud_frame_count, (int)total_aud_frames, samples);
                first_aud = 0;
                long bytes_done_unto = ftell(infile) - audio_started;
                if (bytes_done_unto > frame_size)
                {
                    fprintf(stderr, "processed 0x%lx bytes, should have done 0x%"PRIx32"\n",
                        bytes_done_unto, frame_size);
                    exit(EXIT_FAILURE);
                }
                else if (bytes_done_unto < frame_size)
                {
                    while (bytes_done_unto < frame_size)
                    {
                        get8(infile);
                        bytes_done_unto ++;
                    }
                }
            }
            else
            {
                fprintf(stderr, "unexpected frame id at %08lx\n", (unsigned long)ftell(infile));
                exit(EXIT_FAILURE);
            }
        }

        if (expected_aud_frame_count != aud_frame_count ||
            expected_vid_frame_count != vid_frame_count)
        {
            fprintf(stderr, "frame count mismatch\n");
            exit(EXIT_FAILURE);
        }

        //printf("block %d ended at 0x%lx (%d samples)\n", (int)block_count, ftell(infile), block_sample_count);
        if (ftell(infile) != (data_start+expected_block_size))
        {
            fprintf(stderr, "block size mismatch\n");
            exit(EXIT_FAILURE);
        }
        total_sample_count += block_sample_count;
    }

    if (total_aud_frames != header.audio_frames ||
        total_vid_frames != header.video_frames)
    {
        fprintf(stderr, "total frame count mismatch\n");
        exit(EXIT_FAILURE);
    }

    printf("%"PRIu32" samples\n", total_sample_count);

    // generate header
    make_wav_header(riff_header, total_sample_count, header.audio_srate, 2);
    fseek(outfile, 0, SEEK_SET);
    if (0x2c != fwrite(riff_header, 1, 0x2c, outfile))
    {
        fprintf(stderr, "error rewriting riff header\n");
        exit(EXIT_FAILURE);
    }
    if (EOF == fclose(outfile))
    {
        fprintf(stderr, "error finishing output\n");
        exit(EXIT_FAILURE);
    }

    printf("Done!\n");
}
